{-# LANGUAGE LambdaCase #-}

module PUBMED.Parser where

import Control.Lens ((^?), (^.), (^..), only, to, ix, prism', Prism')
import Data.Maybe (catMaybes, fromMaybe)
import Data.Time (UTCTime(..))
import Data.Time.Segment (jour)
import Prelude hiding (head)
import Text.Taggy (Node(..))
import qualified Data.ByteString.Lazy as DBL
import qualified Data.Text as T
import qualified Data.Text.Lazy as TL
import qualified Data.Text.Lazy.IO as TLIO
import Panic (panic)
import qualified Text.Read as TR
import qualified Text.Taggy.Lens as TTL

namedEl name = TTL.elements . TTL.named (only name) 

contentWithChildren :: Prism' Node T.Text
contentWithChildren = prism' NodeContent $ \case { NodeContent c -> Just c
                                                 ; NodeElement e -> Just $ e ^. TTL.children . traverse . contentWithChildren }

deepContent = TTL.children . traverse . contentWithChildren

parseDocIds :: TL.Text -> [Integer]
parseDocIds txt = map parseId parsed
  where
    parsed = txt ^.. TTL.html . TTL.allNamed (only "eSearchResult") . namedEl "IdList" . namedEl "Id" . TTL.contents
    parseId s = case (TR.readMaybe (T.unpack s) :: Maybe Integer) of
      Nothing -> panic $ "Can't read doc id from: " <> (T.unpack s)
      Just cnt -> cnt

parseDocCount :: TL.Text -> Maybe Integer
parseDocCount txt = TR.readMaybe $ T.unpack $ txt ^. TTL.html . TTL.allNamed (only "eSearchResult") . namedEl "Count" . TTL.contents

data PubMed =
     PubMed { pubmed_article :: PubMedArticle
            , pubmed_date    :: PubMedDate
     } deriving Show

data PubMedArticle =
     PubMedArticle { pubmed_title   :: Maybe T.Text
                   , pubmed_journal :: Maybe T.Text
                   , pubmed_abstract :: [T.Text]
                   , pubmed_authors :: [Author]
                   }
                   deriving (Show)

data Author =
  Author { lastName :: Maybe T.Text
         , foreName :: Maybe T.Text
         , affiliation :: Maybe T.Text
         }
  deriving (Show)

data PubMedDate =
     PubMedDate { pubmedDate_date     :: UTCTime
                , pubmedDate_year     :: Integer
                , pubmedDate_month    :: Int
                , pubmedDate_day      :: Int
                } deriving (Show)


parsePubMedFile fp = do
  input <- TLIO.readFile fp
  pure $ parsePubMed input

parsePubMed :: TL.Text -> [PubMed]
parsePubMed txt = catMaybes $ txt ^.. pubmedArticle . to pubMed
  where
    pubmedArticle = TTL.html . TTL.allNamed (only "PubmedArticleSet") . TTL.allNamed (only "PubmedArticle")
    pubMed el = PubMed <$> el ^? article . to pubMedArticle
                       <*> el ^? pubDate . to pubMedDate
      where
        article = namedEl "MedlineCitation" . namedEl "Article"
        pubDate = namedEl "PubmedData" . namedEl "History" . namedEl "PubMedPubDate" . TTL.attributed (ix "PubStatus" . only "pubmed")
    pubMedDate el = PubMedDate { pubmedDate_date = jour y m d
                               , pubmedDate_year = y
                               , pubmedDate_month = m
                               , pubmedDate_day = d }
      where
        y = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Year" . TTL.contents
        m = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Month" . TTL.contents
        d = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Day" . TTL.contents
    pubMedArticle el = PubMedArticle { pubmed_title = Just $ el ^. title
                                     , pubmed_journal = el ^? journalTitle
                                     , pubmed_abstract = el ^.. abstract
                                     , pubmed_authors = el ^.. authors . to author }
      where
        journalTitle = namedEl "Journal" . namedEl "Title" . TTL.contents
        title = namedEl "ArticleTitle" . deepContent
        abstract = namedEl "Abstract" . TTL.allNamed (only "AbstractText") . to abstractContent
          where
            abstractContent el = el ^. deepContent
        authors = namedEl "AuthorList" . TTL.allNamed (only "Author")
    author el = Author { lastName = el ^? lastName
                       , foreName = el ^? firstName
                       , affiliation = el ^? affiliation }
      where
        firstName = namedEl "ForeName" . TTL.contents
        lastName = namedEl "LastName" . TTL.contents
        affiliation = namedEl "AffiliationInfo" . namedEl "Affiliation" . TTL.contents

--identity :: a -> a
--identity x = x
--
--manyTagsUntil :: MonadThrow m =>
--     Name
--  -> ConduitT Event o m b
--  -> ConduitT Event o m (Maybe b)
--manyTagsUntil name f = do
--  _ <- manyTagsUntil_ name
--  tagIgnoreAttrs (matching (== name)) f
--
---- | Utility function that matches everything but the tag given
--tagUntil :: Name -> NameMatcher Name
--tagUntil name = matching (/= name)
--
---- | Utility function that consumes everything but the tag given
---- usefull because we have to consume every data.
--manyTagsUntil_ :: MonadThrow m => Name -> ConduitT Event o m ()
--manyTagsUntil_ n = many_ ( ignoreTree (tagUntil n) ignoreAttrs)
--
--manyTagsUntil_' :: MonadThrow m => Name -> ConduitT Event o m ()
--manyTagsUntil_' = many_ . ignoreEmptyTag . tagUntil

--parseDocId :: C.Cursor -> [Integer]
--parseDocId cursor = fst <$>
--  rights (T.decimal
--          <$> filter notNullOrEOL (rawElement cursor)
--         )
--  where rawElement = C.element "eSearchResult" C.&/ C.element "IdList"  C.&// C.content
--        notNullOrEOL t = not (T.null t) && t /= "\n"

--readPubMedFile :: FilePath -> IO [PubMed]
--readPubMedFile fp = do
--  input <- DBL.readFile fp
--  pubMedParser input

--pubMedParser :: DBL.ByteString -> IO [PubMed]
--pubMedParser bstring = runConduit $ parseLBS def bstring
--                                 .| parseArticleSet
--                                 .| CL.consume
--
--parseArticleSet :: MonadThrow m => ConduitT Event PubMed m ()
--parseArticleSet =
--  force "PubmedArticleSet required" $ manyTagsUntil "PubmedArticleSet" $ manyYield parsePubMedArticle
--
--parsePubMedArticle :: MonadThrow m => ConduitT Event o m (Maybe PubMed)
--parsePubMedArticle =
--  manyTagsUntil "PubmedArticle" parsePubMedArticle'
--
--parsePubMedArticle' :: MonadThrow m => ConduitT Event o m PubMed
--parsePubMedArticle' = do
--  article <- force "MedlineCitation" $ tagIgnoreAttrs "MedlineCitation" parseMedlineCitation
--  dates <- tagIgnoreAttrs "PubmedData" $ do
--    dates' <- tagIgnoreAttrs "History" $ many $ tagIgnoreAttrs "PubMedPubDate" $ do
--      y' <- force "Year"  $ tagIgnoreAttrs "Year"  content
--      m' <- force "Month" $ tagIgnoreAttrs "Month" content
--      d' <- force "Day"   $ tagIgnoreAttrs "Day"   content
--      _ <- many ignoreAnyTreeContent
--      return (read $ unpack y', read $ unpack m', read $ unpack d')
--    _ <- many ignoreAnyTreeContent
--    return dates'
--  _ <- many ignoreAnyTreeContent
--  let (y,m,d) = maybe (1,1,1) (fromJust . head) (reverse <$> join dates)
--  return $ PubMed article (PubMedDate (jour y m d) y m d)
--
--parseMedlineCitation :: MonadThrow m => ConduitT Event o m PubMedArticle
--parseMedlineCitation = do
--  a <- force "article" $ manyTagsUntil "Article" parseArticle
--  _ <- many ignoreAnyTreeContent
--  return a
--
--parseArticle :: MonadThrow m => ConduitT Event o m PubMedArticle
--parseArticle = do
--  journal  <- force "journal" $
--    manyTagsUntil "Journal" $ do
--    j <- manyTagsUntil "Title" content
--    _ <- many ignoreAnyTreeContent
--    return j
--
--  title <- manyTagsUntil "ArticleTitle" content
--
--  abstracts <- manyTagsUntil "Abstract" . many $ do
--      txt <- tagIgnoreAttrs "AbstractText" content
--      _ <- many ignoreAnyTreeContent
--      return txt
--      -- TODO add authos
--
--  authors <-  manyTagsUntil "AuthorList" . many $
--     tagIgnoreAttrs "Author" $ do
--        ln <- manyTagsUntil "LastName" content
--        fn <- manyTagsUntil "ForeName" content
--        affi <- manyTagsUntil "AffiliationInfo" $ do
--          aff <- manyTagsUntil "Affiliation" content
--          _ <- many ignoreAnyTreeContent
--          return aff
--        _ <- many ignoreAnyTreeContent
--        return Author {lastName=ln, foreName=fn, affiliation=fromMaybe Nothing affi}
--
--
--  _ <- many ignoreAnyTreeContent
--  return $ PubMedArticle title journal abstracts authors


pubMedData :: DBL.ByteString
pubMedData = mconcat
  [ "<?xml version=\"1.0\"?>\n"
  , "<!DOCTYPE PubmedArticleSet PUBLIC \"-//NLM//DTD PubMedArticle, 1st June 2018//EN\" \"https://dtd.nlm.nih.gov/ncbi/pubmed/out/pubmed_180601.dtd\">\n"
  , "<PubmedArticleSet>\n"
  , "<PubmedArticle>\n"
  , " <MedlineCitation Status=\"Publisher\" Owner=\"NLM\">\n"
  , "        <PMID Version=\"1\">30357468</PMID>\n"
  , "        <DateRevised>\n"
  , "           <Year>2018</Year>\n"
  , "        </DateRevised>\n"
  , "        <Article PubModel=\"Print-Electronic\">\n"
  , "          <Journal>\n"
  , "            <ISSN IssnType=\"Electronic\">1432-1076</ISSN>\n"
  , "            <Title>European journal of pediatrics</Title>\n"
  , "          </Journal>\n"
  , "          <ArticleTitle>Title of the Article</ArticleTitle>\n"
  , "          <ELocationID EIdType=\"doi\" ValidYN=\"Y\">10.1007/s00431-018-3270-3</ELocationID>\n"
  , "          <Abstract>\n"
  , "              <AbstractText>Abstract Text.</AbstractText>\n"
  , "          </Abstract>\n"
  , "          <AuthorList>\n"
  , "          </AuthorList>\n"
  , "        </Article>\n"
  , " </MedlineCitation>\n"
  , " <PubmedData>\n"
  , "   <History>\n"
  , "   </History>\n"
  , " </PubmedData>\n"
  , "</PubmedArticle>\n"
  , "</PubmedArticleSet>\n"
  ]
